
#include <GL/gl.h>
#include <cstdio>
#include "Isosurface.h"


void Octree::marchingTets(unsigned char isovalue, Vector<Point3D> &verts, Vector<Vector3> &normals, int maxTris) {
  root->marchingTets(vol, isovalue, verts, normals, sizeX, sizeY, sizeZ, maxTris);
} // Octree::marchingTets()


void Octree::build(const unsigned char *v, int sizeX, int sizeY, int sizeZ) {
  if ( root )
    delete root;

  // store pointer to volume data
  vol = v;
  this->sizeX = sizeX;
  this->sizeY = sizeY;
  this->sizeZ = sizeZ;

  root = new OctreeNode(vol, 0, 0, 0, sizeX, sizeY, sizeZ, sizeX, sizeY, sizeZ, 0, 4);
} // Octree::build()


OctreeNode::OctreeNode(const unsigned char *vol, int _x, int _y, int _z, int _sx, int _sy, int _sz, int sizeX, int sizeY, int sizeZ, int depth, int maxDepth) {
  min = 255;
  max = 0;

  x = _x;
  y = _y;
  z = _z;

  sx = _sx;
  sy = _sy;
  sz = _sz;

  // x, y, and z sizes of child nodes
  int childSX = sx >> 1,
      childSY = sy >> 1,
      childSZ = sz >> 1;

  bool forceLeaf = false;

  if ( childSX <= 1 || childSY <= 1 || childSZ <= 1 )
    forceLeaf = true;

  if ( depth < maxDepth && !forceLeaf ) {
    int offsetX = (sx % 2) ? 1 : 0, offsetY = (sy % 2) ? 1 : 0, offsetZ = (sz % 2) ? 1 : 0;

    for ( int i=0; i < 2; i++ )
      for ( int j=0; j < 2; j++)
        for ( int k=0; k < 2; k++ ) {
          int index = (k << 2) + (j << 1) + i;
          int offX = ( 1 == i ) ? offsetX : 0;
          int offY = ( 1 == j ) ? offsetY : 0;
          int offZ = ( 1 == k ) ? offsetZ : 0;
          int szOffX = ( 0 == i ) ? offsetX : 0;
          int szOffY = ( 0 == j ) ? offsetY : 0;
          int szOffZ = ( 0 == k ) ? offsetZ : 0;

          children[index] = new OctreeNode(vol, x + i*childSX + offX, y + j*childSY + offY, z + k*childSZ + offZ, childSX + szOffX, childSY + szOffY, childSZ + szOffZ, sizeX, sizeY, sizeZ, depth+1, maxDepth);
        }
    // find min and max in child nodes
    for (int i=0; i < 8; i++) {
      unsigned char childMin = children[i]->min, childMax = children[i]->max;
      if ( childMin < min )
        min = childMin;
      if ( childMax > max )
        max = childMax;
    }

  } // if
  else { // leaf node

  int endX = x+sx >= sizeX ? x+sx : x+sx+1,
      endY = y+sy >= sizeY ? y+sy : y+sy+1,
      endZ = z+sz >= sizeZ ? z+sz : z+sz+1;

  int sizeXY = sizeX*sizeY;
    // find min and max
    for (int i=x; i < endX; i++)
      for (int j=y; j < endY; j++)
        for (int k=z; k < endZ; k++) {
          unsigned char vxl = vol[k*sizeXY + j*sizeX + i];
          if ( vxl > max )
            max = vxl;
          if ( vxl < min )
            min = vxl;
        }

    // set children to NULL
    for (int i=0; i < 8; i++)
      children[i] = 0;
  } // else

} // OctreeNode::OctreeNode()


OctreeNode::~OctreeNode() {
  for (int i=0; i < 8; i++)
    if ( children[i] )
      delete children[i];
} // OctreeNode::~OctreeNode()


void OctreeNode::marchingTets(const unsigned char *vol, unsigned char isovalue, Vector<Point3D> &verts, Vector<Vector3> &normals,
                              int sizeX, int sizeY, int sizeZ, int maxTris) {
  if ( (min < isovalue && max < isovalue) ||
       (min > isovalue && max > isovalue) ||
       min == max )
    return;

  if ( children[0] ) { // not a leaf
    for (int i=0; i < 8; i++) {
      children[i]->marchingTets(vol, isovalue, verts, normals, sizeX, sizeY, sizeZ, maxTris);

      // if passed triangle limit, return
      if ( (verts.getSize() >= maxTris*3 && maxTris != 0) || verts.isFull() || normals.isFull() )
        return;
    }
  }
  else { // a leaf
    int endX = x+sx >= sizeX ? sx-1 : sx,
	endY = y+sy >= sizeY ? sy-1 : sy,
	endZ = z+sz >= sizeZ ? sz-1 : sz;

    int maxVerts = maxTris*3;    

    for (int i=0; i < endX; i++) {
      for (int j=0; j < endY; j++) {
        for (int k=0; k < endZ; k++) {
          // if passed triangle limit, return
	  if ( (verts.getSize() >= maxVerts && maxVerts != 0) || verts.isFull() || normals.isFull() )
            return;

          Isosurface::marchTet(verts, normals, vol, isovalue, x+i, y+j, z+k, sizeX, sizeY, sizeZ);
        } // for k
      } // for j
    } // for i

  } // else

} // OctreeNode::marchingTets()


void Isosurface::setVol(const unsigned char *v, int sx, int sy, int sz) {
  sizeX = sx;
  sizeY = sy;
  sizeZ = sz;

  dimX = (float)sizeX;
  dimY = (float)sizeY;
  dimZ = (float)sizeZ;

  float longest = dimX > dimY ? dimX : dimY;
  longest = longest > dimZ ? longest : dimZ;

  scaleX = scaleY = scaleZ = 1.0f/longest;

  vol = v;

  octree.build(v, sizeX, sizeY, sizeZ);
} // Isosurface::setVol()


// constructs an isosurface from volume data (a scalar field) using the
// marching tetrahedra algorithm
bool Isosurface::marchingTets(unsigned char isovalue, int maxTris) {
  // sanity check
  if ( !vol )
    return true;

  // clear out any data that may be loaded...
  verts.clear();
  normals.clear();

  octree.marchingTets(isovalue, verts, normals, maxTris);

  // return false if max # of triangles was exceeded
  if ( verts.isFull() || normals.isFull() ||
       (maxTris > 0 && verts.getSize()*3 == maxTris) )
    return false;

  return true;
} // Isosurface::marchingTets()


inline Vector3 computeGradient(const unsigned char *vol, int x, int y, int z, int sizeX, int sizeY, int sizeZ) {
  int right   = x + 1 >= sizeX ? sizeX-1 : x + 1,
      left    = x - 1 <= 0 ? 0 : x - 1,
      up      = y + 1 >= sizeY ? sizeY-1 : y + 1,
      down    = y - 1 <= 0 ? 0 : y - 1,
      forward = z + 1 >= sizeZ ? sizeZ-1 : z + 1,
      back    = z - 1 <= 0 ? 0 : z - 1;

   int sizeXY = sizeX*sizeY;

   float s0 = (float)vol[z*sizeXY + y*sizeX + right],
         s1 = (float)vol[z*sizeXY + y*sizeX + left],
         s2 = (float)vol[z*sizeXY + up*sizeX + x],
         s3 = (float)vol[z*sizeXY + down*sizeX + x],
         s4 = (float)vol[forward*sizeXY + y*sizeX + x],
         s5 = (float)vol[back*sizeXY + y*sizeX + x];

   Vector3 gradient(s0 - s1, s2 - s3, s4 - s5);
   //gradient *= 0.5f; // supposidly gives better results...
   gradient.normalize();

   return gradient;
} // computeGradient()


inline Vector3 computeGradient(unsigned char left, unsigned char right, unsigned char up, unsigned char down,
                               unsigned char forward, unsigned char back) {
   float s0 = (float)left,
         s1 = (float)right,
         s2 = (float)up,
         s3 = (float)down,
         s4 = (float)forward,
         s5 = (float)back;

   Vector3 gradient(s0 - s1, s2 - s3, s4 - s5);
   gradient *= 0.5f; // supposidly gives better results...
   //gradient.normalize();

   return gradient;
} // computeGradient()


inline Vector3 computeTriNormal(const Point3D &v0, const Point3D &v1, const Point3D &v2) {
  Vector3 e0 = v0 - v2, e1 = v0 - v1;

  // TODO: maybe i can take this out
  //e0.normalize();
  //e1.normalize();

  Vector3 normal;
  normal.cross(e0, e1);
  normal.normalize();

  return normal;
} // computeTriNormal()


void Isosurface::marchTet(Vector<Point3D> &verts, Vector<Vector3> &normals, const unsigned char *vol, 
                          unsigned char isovalue, int x, int y, int z, int sizeX, int sizeY, int sizeZ) {

  static const int tetI[6][4] = {  { 0, 1, 2, 5 },
                                   { 1, 3, 5, 6 },
                                   { 1, 2, 3, 5 },
                                   { 1, 5, 6, 7 },
                                   { 4, 5, 7, 1 },
                                   { 1, 5, 0, 4 } };

  static const Point3D cube[8] = { Point3D(0.0f, 0.0f, 0.0f),
                                   Point3D(1.0f, 0.0f, 0.0f),
                                   Point3D(0.0f, 0.0f, 1.0f),
                                   Point3D(1.0f, 0.0f, 1.0f),
                                   Point3D(0.0f, 1.0f, 0.0f),
                                   Point3D(0.0f, 1.0f, 1.0f),
                                   Point3D(1.0f, 1.0f, 1.0f),
                                   Point3D(1.0f, 1.0f, 0.0f) };

  int sizeXY = sizeX*sizeY, ySizeX = y*sizeX;
  // extract isovalues from vertices of cube
  unsigned char iso[8];
  iso[0] = vol[z*sizeXY     + ySizeX      + x  ]; // bottom left front
  iso[1] = vol[z*sizeXY     + ySizeX      + x+1]; // bottom right front
  iso[2] = vol[(z+1)*sizeXY + ySizeX      + x  ]; // top left front
  iso[3] = vol[(z+1)*sizeXY + ySizeX      + x+1]; // bottom left back
  iso[4] = vol[z*sizeXY     + (y+1)*sizeX + x  ]; // top right front
  iso[5] = vol[(z+1)*sizeXY + (y+1)*sizeX + x  ]; // bottom right back
  iso[6] = vol[(z+1)*sizeXY + (y+1)*sizeX + x+1]; // top left back
  iso[7] = vol[z*sizeXY     + (y+1)*sizeX + x+1]; // top right back

  // bit flags to indicate which vertices are inside the surface
  // (greater than isovalue)
  unsigned char inside = 0;
  for (int i=0; i < 8; i++)
    if ( iso[i] < isovalue )
      inside |= 1 << i; // inside the surface

  // completely inside or comletely outside?
  if ( !inside || 255 == inside )
    return;

  Vector3 gradients[8] = { computeGradient(vol, x, y, z, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x+1, y, z, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x, y, z+1, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x+1, y, z+1, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x, y+1, z, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x, y+1, z+1, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x+1, y+1, z+1, sizeX, sizeY, sizeZ),
                           computeGradient(vol, x+1, y+1, z, sizeX, sizeY, sizeZ) };

  // corner of voxel
  Point3D p0( (float)x, (float)y, (float)z );

  for (int i=0; i < 6; i++) {

    int count = 0, insideIndex[4], outsideIndex[4];
    for (int j=0; j < 4; j++)
      if ( inside & (1 << tetI[i][j]) )
        insideIndex[count++] = tetI[i][j];
      else
        outsideIndex[j - count] = tetI[i][j];

    if (2 == count) {
      int i0 = insideIndex[0], i1 = insideIndex[1],
          o0 = outsideIndex[0], o1 = outsideIndex[1];

      Point3D v0 = p0 + cube[o0] + (cube[i0] - cube[o0])*0.5f,
              v1 = p0 + cube[o0] + (cube[i1] - cube[o0])*0.5f,
              v2 = p0 + cube[o1] + (cube[i0] - cube[o1])*0.5f,
              v3 = p0 + cube[o1] + (cube[i1] - cube[o1])*0.5f;

      Vector3 n0 = gradients[i0] + gradients[o0],
              n1 = gradients[i1] + gradients[o0],
              n2 = gradients[i0] + gradients[o1],
              n3 = gradients[i1] + gradients[o1];

      n0.normalize();
      n1.normalize();
      n2.normalize();
      n3.normalize();

      // add to vertex array, and handle
      // overflow exceptions  
      verts.insert(v0);
      verts.insert(v1);
      verts.insert(v2);
      verts.insert(v2);
      verts.insert(v1);
      verts.insert(v3);

      normals.insert(n0);
      normals.insert(n1);
      normals.insert(n2);
      normals.insert(n2);
      normals.insert(n1);
      normals.insert(n3);
    } // if ( 2 == count )
    else if (1 == count || 3 == count) {
      int vIndex = 0;
      Point3D v[3];
      Vector3 n[3];

      for (int k=0; k < 4; k++)
        for (int l=k+1; l < 4; l++) {
          bool b0 = iso[ tetI[i][k] ] < isovalue, b1 = iso[ tetI[i][l] ] < isovalue;
          if ( b0 != b1 ) {
            n[vIndex] = gradients[tetI[i][l]] + gradients[tetI[i][k]];
            v[vIndex++] = p0 + cube[tetI[i][k]] + (cube[tetI[i][l]] - cube[tetI[i][k]])*0.5f;
          }
        }

      // add to vertex array, and handle
      // overflow exceptions
      for (int k=0; k < 3; k++) {
        n[k].normalize();
        normals.insert(n[k]);
        verts.insert(v[k]);
      }
    } // else-if

  } // for

} // Isosurface::marchTet()


void Isosurface::render() const {
  if ( verts.getSize() == 0 )
    return;

  glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
 
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);

  glVertexPointer( 3, GL_FLOAT, 0, verts.getData() );
  glNormalPointer( GL_FLOAT, 0, normals.getData() );

  glPushMatrix();
    glScalef(scaleX, scaleY, -scaleZ);
    glTranslatef(dimX*-0.5f, dimY*-0.5f, dimZ*-0.5f);
    glDrawArrays( GL_TRIANGLES, 0, verts.getSize() );
  glPopMatrix();

  glPopClientAttrib();
} // Isosurface::clear()


void Isosurface::clear() {
  verts.clear();
  normals.clear();
} // Isosurface::clear()

