
#ifndef VECTOR3_H
#define VECTOR3_H

#include <cmath>

class Vector3 {
public:
   // components are stored in this union so they can be accessed as an array,
   // or by component name (x, y, z).
   union {
      float m[3];
      struct { float x, y, z; };
   }; // union

   static const Vector3 zero;

   Vector3() { }
   Vector3(float _x, float _y, float _z): x(_x), y(_y), z(_z) { }

   // overloaded operators
   inline float & operator [] (int i);
   inline Vector3 operator + (const Vector3 &v) const;
   inline const Vector3 & operator += (const Vector3 &v);
   inline Vector3 operator - (const Vector3 &v) const;
   inline const Vector3 & operator -= (const Vector3 &v);
   inline float operator * (const Vector3 &v) const;
   inline Vector3 operator * (float s) const;
   inline const Vector3 & operator *= (float s);

   // functions
   inline void cross(const Vector3 &u, const Vector3 &v);
   inline float mag() const;
   inline void normalize();
   inline Vector3 reflect(const Vector3 &n) const;
   inline void reflect(const Vector3 &n, Vector3 &result);
   inline void set(float _x, float _y, float _z);
   inline float sqrMag() const;
};

inline float & Vector3::operator [] (int i) {
   return m[i];
} // Vector3::operator []

// This overloaded + operator performs vector addition.  The Vector3 object is added
// to another Vector3 object, v, and the result is returned.
inline Vector3 Vector3::operator + (const Vector3 &v) const {
   return Vector3(x + v.x, y + v.y, z + v.z);
} // Vector3::operator +

// This overloaded += operator performs vector addition and stores the result in the
// Vector3 object.
inline const Vector3 & Vector3::operator += (const Vector3 &v) {
   x += v.x;
   y += v.y;
   z += v.z;

   return (*this);
} // Vector3::operator +=

// This overloaded - operator performs vector subtraction.  v is subtracted from the Vector3
// object, and the result is returned.
inline Vector3 Vector3::operator - (const Vector3 &v) const {
   return Vector3(x - v.x, y - v.y, z - v.z);
} // Vector3::operator -

// performs vector subtraction, subtracts v from the Vector3 object and stores the result in
// itself.
inline const Vector3 & Vector3::operator -= (const Vector3 &v) {
   x -= v.x;
   y -= v.y;
   z -= v.z;

   return (*this);
} // Vector3::operator -=

// overloaded * operator for scalar multiplication, multiplies this Vector3 object by
// the scalar and returns the resulting Vector3.
inline Vector3 Vector3::operator * (float s) const {
   return Vector3(x*s, y*s, z*s);
} // Vector3::operator *

// multiplies Vector3 by the scalar, and stores the result in itself.
inline const Vector3 & Vector3::operator *= (float s) {
   x *= s;
   y *= s;
   z *= s;

   return (*this);
} // Vector3::operator *=

// This overloaded * operator returns the dot product of the vector, and another vector v.
inline float Vector3::operator *(const Vector3 &v) const {
   return x*v.x + y*v.y + z*v.z;
} // Vector3::operator *()

// This function computes the cross product of two vectors and stores the result
// in the Vector3 object.
inline void Vector3::cross(const Vector3 &u, const Vector3 &v) {
   x = u.y*v.z - u.z*v.y;     
   y = u.z*v.x - u.x*v.z;
   z = u.x*v.y - u.y*v.x;
} // Vector3::cross()

// returns the magnitude of the vector
inline float Vector3::mag() const {
   return sqrtf(x*x + y*y + z*z);
} // Vector3::mag()

// normalizes the vector
inline void Vector3::normalize() {
   float len = sqrtf(x*x + y*y + z*z);

   x /= len;
   y /= len;
   z /= len;
} // Vector3::normalize()

// reflect vector about vector n and return result
inline Vector3 Vector3::reflect(const Vector3 &n) const {
   return n*2.0f*( n*(*this) ) - (*this);
} // Vector3::reflect()

// stores vector reflected about n in 'result'
inline void Vector3::reflect(const Vector3 &n, Vector3 &result) {
   result = n*2.0f*( n*(*this) ) - (*this);
} // Vector3::reflect()

// returns the squared magnitude of the vector (this is faster than mag(), since it
// does not involve the computation of a square root)
inline float Vector3::sqrMag() const {
   return x*x + y*y + z*z;
} // Vector3::mag()

// assign the values of the vector's x, y, and z components to _x, _y, and _z respectively.
inline void Vector3::set(float _x, float _y, float _z) {
   x = _x;
   y = _y;
   z = _z;
} // Vector3::set()


// scalar-vector multiplication (with scalar on left side)
inline Vector3 operator * (float s, const Vector3 &v) {
  return Vector3(s*v.x, s*v.y, s*v.z);
} // operator *

#endif

