
#include <cstdlib>
#include "VisCanvas.h"


VisCanvas::VisCanvas(QWidget *parent):
     QGLWidget(parent) {
  // enable mouse tracking for this widget so trackball
  // can be used
  setMouseTracking(true);
  zoom = -5.0f;
  vol = NULL;
  isovalue = 40;
  stepSize = 0.1f;
  renderMode = RENDER_ISOSURFACE;

  perspProj = true;

  // setup view frustum properties
  vFOV = 45.0f;
  left = -1.5f;
  right = 1.5f;
  top = 1.5f;
  bottom = 1.5f;
  nearZ = 0.1f;
  farZ = 100.0f;
} // VisCanvas::VisCanvas()


VisCanvas::~VisCanvas() {
  if ( vol )
    delete vol;
} // VisCanvas::~VisCanvas()


// loads a volume file
bool VisCanvas::loadVolume(const char *filename) {
  // sanity check
  if ( !filename )
    return false;

  Vol *newVol = new Vol(filename);

  // was load successful
  if ( !newVol->isOK() ) {
    delete newVol;
    newVol = NULL;
    return false;
  }

  // if there is already volume data, deallocate it
  if ( vol )
    delete vol;

  vol = newVol;
  volumeScene.setVol( vol->getData(), vol->sizeX(), vol->sizeY(), vol->sizeZ() );

  render();

  // success
  return true;
} // VisCanvas::loadVolume()


// inherited from QGLWidget, called once before the first call to
// paintGL() or resizeGL(), and then it is called whenever the widget
// is assigned a new QGLContext
void VisCanvas::initializeGL() {
  glLoadIdentity();

  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
  glEnable(GL_NORMALIZE); // because i use scaling...

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  float lightPos[] = { 0.0f, -1.0f, 0.0f, 0.0f };
  float lightDiffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  float lightAmbient[] = { 0.5f, 0.5f, 0.5f, 1.0f };
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
} // VisCanvas::initializeGL()


// inherited from QGLWidget, called when widget is resized.
// sets viewport and projection matrix
void VisCanvas::resizeGL(int w, int h) {
  // make sure width and height are at least 1
  if ( w <= 0 )
    w = 1;
  if ( h <= 0 )
    h = 1;

  // set viewport
  glViewport(0, 0, w, h);

  // set projection matrix
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  // TODO: check setting!
  // perspective projection
  gluPerspective(vFOV, ((float)w)/((float)h), nearZ, farZ);
  //glOrtho(-1.5f, 1.5f, -1.5f, 1.5f, 0.0f, 10.0f);

  glMatrixMode(GL_MODELVIEW);

  volumeScene.resize(w, h);
  //if ( !volumeScene.resize(w, h) )
    //close();

  render();
} // VisCanvas::resizeGL()


// inherited from QGLWidget, called when window
// needs to be updated (repainted)
void VisCanvas::paintGL() {
  volumeScene.display();
} // VisCanvas::paintGL()


// called when a new image needs to be generated
void VisCanvas::render() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // set modelview matrix to identity matrix
  glLoadIdentity();

  // translate to camera position
  glTranslatef(0.0f, 0.0f, zoom);

  // get rotation matrix from trackball
  float mat[16];
  trackball.look(mat);
  glMultMatrixf(mat);

  switch(renderMode) {
      case RENDER_SLICE:  /* not implemented */ break;
      case RENDER_ISOSURFACE: volumeScene.renderIso(); break;
      case RENDER_VOLUME: /* not implemented */ break;
  }

  glFlush();
  updateGL();
} // VisCanvas::render()


// called when user moves mouse
void VisCanvas::mouseMoveEvent(QMouseEvent *ev) {
  // trackball can only be moved when trackball is enabled, but check
  // to see if it's enabled anyway to prevent an unnecessary repaint.
  if ( trackball.isTracking() ) {
    trackball.move( ev->x(), ev->y(), width(), height() );
    render();
  }
} // VisCanvas::mouseMoveEvent()


// event handler called when user presses mouse button over
// the widget
void VisCanvas::mousePressEvent(QMouseEvent *ev) {
  // update trackball if user has pressed the left mouse button
  Qt::MouseButtons buttons = ev->buttons();
  if ( buttons & Qt::LeftButton ) {
    // enable tracking if it isn't enabled already
    if ( !trackball.isTracking() ) {
      trackball.startTracking();
      trackball.startMotion( ev->x(), ev->y(), width(), height() );
    }

    render();
  }
} // VisCanvas::mousePressEvent()


// called when user releases a mouse button
void VisCanvas::mouseReleaseEvent(QMouseEvent *ev) {
  // if user has released the left mouse button, stop trackball
  // from tracking mouse motion
  Qt::MouseButtons buttons = ev->buttons();
  if ( !(buttons & Qt::LeftButton ) ) {
    trackball.stopTracking();
    render();
  }
} // VisCanvas::mouseReleaseEvent()


// event handler called when user uses mouse wheel and this widget
// is selected
void VisCanvas::wheelEvent(QWheelEvent *ev) {
  // accept() indicates that we don't want the message to be
  // sent to the parent widget
  ev->accept();

  int delta = ev->delta();

  if ( delta > 0 )
    zoom -= 0.5f;
  else if ( delta < 0 )
    zoom += 0.5f;

  render();
} // VisCanvas::wheelEvent()


// event handler called when user presses a key and this widget
// is selected
void VisCanvas::keyPressEvent(QKeyEvent *ev) {
  if ( ev->key() == Qt::Key_A )
    zoom -= 0.5f;
  else if ( ev->key() == Qt::Key_B )
    zoom += 0.5f;

  render();
} // VisCanvas::keyPressEvent()


// called when user presses "build isosurface"
// button
void VisCanvas::buildIsoSlot() {
  if ( !volumeScene.buildIso(isovalue, maxTris) )
    QMessageBox::warning(this, "Too many triangles!", "Either the maximum number"
                               " of triangle was exceeded or your system could "
                               " not support the number of triangles generated.");

  render();
} // VisCanvas::buildIsoSlot()


// called when user changed maximum # of triangles
void VisCanvas::maxTriSlot(const QString &s) {
  maxTris = s.toInt();
} // VisCanvas::maxTriSlot()

/*
// slot for orthographic projection
// radio button
void VisCanvas::orthSlot(bool on) {
  if ( on ) {
    // set projection matrix to perform orthographic projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glOrtho(left, right, top, bottom, nearZ, farZ);
    glMatrixMode(GL_MODELVIEW);
  }

  render();
} // VisCanvas::orthoSlot


// slot for perspective projection
// radio button
void VisCanvas::perspSlot(bool on) {
  if ( on ) {
    // set projection matrix to perform perspective projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    gluPerspective(vFOV, ((float)width())/((float)height()), nearZ, farZ);
    
    glMatrixMode(GL_MODELVIEW);
  }

  perspProj = on;

  render();
} // VisCanvas::perspSlot()
*/

void VisCanvas::isovalueChangedSlot(int value) {
  isovalue = (unsigned char)value;
} // VisCanvas::isovalueChangedSlot()

