
#include "trackball.h"


// move() should be called whenever the mouse is moving
void Trackball::move(int x, int y, int w, int h) {
   if ( !tracking )
      return;

   Vector3 currPos, d;

   // map x, y onto hemi-sphere
   pToV(x, y, w, h, currPos);

   // find change in position since last call
   d = currPos - lastPos;

   // create a quaternion to represent the rotation
   Quat currQuat;

   if ( fabsf(d.x) > 0.01f || fabs(d.y) > 0.01f || fabsf(d.z) > 0.01f ) {
      // compute angle
      float angle = acosf( lastPos*currPos/(lastPos.mag()*currPos.mag()) );
      
      // convert angle from radians to degrees
      angle *= 180.0f/3.14f;

      // calculate axis to rotate about from last position vector and
      // current position vector
      Vector3 axis;
      axis.cross(lastPos, currPos);

      // set last position
      lastPos = currPos;

      // normalize axis of rotation
      axis.normalize();   

      // build a quaternion from the axis, angle pair
      currQuat.buildFromAxisAngle(axis, angle);

      // store all rotations up to this point in q
      q = currQuat*q;      
   } // if
} // Trackball::move()


// stores trackball's transformation matrix in mat.
void Trackball::look(float *mat) {   
   q.toMatrix(mat);   
} // Trackball::look()


// project cursor position (x and y) onto trackball hemi-sphere
void Trackball::pToV(int x, int y, int w, int h, Vector3 &p) {
   float fX = (float)x, fY = (float)y,
         fW = (float)w, fH = (float)h;

   p.x = (2.0f*fX - fW)/fW;
   p.y = (fH - 2.0f*fY)/fH;

   // find distance from origin to selected x, y
   float d = sqrtf(p.x*p.x + p.y*p.y);

   // Find height of selected point on hemi-sphere.  If
   // the cursor is outside of the hemi-sphere (distance of
   // selected x, y from origin is greater than 1) just use
   // a distance of 1.  The cosine of this distance*PI/2 gives
   // the height of the selected point.
   p.z = cosf( (PI/2.0f) * ( (d < 1.0f) ? d : 1.0f) );
   
   p.normalize();
} // Trackball::pToV()


// This function should be called when you are starting
// to move the trackball.  x and y are the current cursor
// coordinates. w and h are the window dimensions.
void Trackball::startMotion(int x, int y, int w, int h) {
   if ( !tracking )
      return;
   pToV(x, y, w, h, lastPos);   
} // Trackball::startMotion()


// multiplies two quaternions
Quat Quat::operator *(const Quat &q) const {
   Quat r;

   r.w = w*q.w - x*q.x - y*q.y - z*q.z;
	r.x = w*q.x + x*q.w + y*q.z - z*q.y;
	r.y = w*q.y + y*q.w + z*q.x - x*q.z;
	r.z = w*q.z + z*q.w + x*q.y - y*q.x;

   return r;
} // Quat::operator *


// sets the value of a quaternion to
// the result of itself multiplied by
// another quaternion, q
const Quat& Quat::operator *=(const Quat &q) {
   // temporary values
   float tx, ty, tz, tw;

   tw = w*q.w - x*q.x - y*q.y - z*q.z;
   tx = w*q.x + x*q.w + y*q.z - z*q.y;
   ty = w*q.y + y*q.w + z*q.x - x*q.z;
   tz = w*q.z + z*q.w + x*q.y - y*q.x;
   
   x=tx; y=ty; z=tz; w=tw;

   return (*this);
} // Quat::opertor *=


// convert quaternion to a rotation matrix 
void Quat::toMatrix(float matrix[]) {
  // column 1
  matrix[ 0] = 1.0f - 2.0f * ( y * y + z * z ); 
	matrix[ 1] = 2.0f * (x * y + z * w);
	matrix[ 2] = 2.0f * (x * z - y * w);
	matrix[ 3] = 0.0f;  

	// column 2
	matrix[ 4] = 2.0f * ( x * y - z * w );  
	matrix[ 5] = 1.0f - 2.0f * ( x * x + z * z ); 
	matrix[ 6] = 2.0f * (z * y + x * w );  
	matrix[ 7] = 0.0f;  

	// column 3
	matrix[ 8] = 2.0f * ( x * z + y * w );
	matrix[ 9] = 2.0f * ( y * z - x * w );
	matrix[10] = 1.0f - 2.0f * ( x * x + y * y );  
	matrix[11] = 0.0f;  

	// column 4
	matrix[12] = 0;  
	matrix[13] = 0;  
	matrix[14] = 0;  
	matrix[15] = 1.0f;
} // Quat::toMatrix()


// construct quaternion from an axis/angle pair
void Quat::buildFromAxisAngle(const Vector3 &v, float angle) {
   // convert to radians (angle should be in degrees)
   float radians = (angle/180.0f)*PI;

   // cache this, since it is used multiple times below
   float sinThetaDiv2 = sinf(radians/2.0f);

   // now calculate the components of the quaternion	
	x = v.x * sinThetaDiv2;
	y = v.y * sinThetaDiv2;
	z = v.z * sinThetaDiv2;

   w = cosf( radians / 2.0f );
} // Quat::buildFromAxisAngle()


